import pytest
import torch

@pytest.mark.parametrize('groups', (1, 4))
def test_mlp(groups):
    from x_mlps_pytorch.grouped_mlp import GroupedMLP

    mlp = GroupedMLP(256, 128, 64, groups = groups)

    x = torch.randn(7, 3, 256)

    assert mlp(x).shape == (7, 3, groups, 64)

# with depth

@pytest.mark.parametrize('groups', (1, 4))
def test_create_mlp(groups):
    from x_mlps_pytorch.grouped_mlp import create_grouped_mlp

    mlp = create_grouped_mlp(
        dim = 128,
        dim_in = 256,
        dim_out = 64,
        depth = 4,
        groups = groups
    )

    # same as GroupedMLP(256, 128, 128, 128, 128, 64)

    x = torch.randn(7, 3, 256)

    assert mlp(x).shape == (7, 3, groups, 64)

# test auto squeeze 1 group, so it can act as regular MLP

def test_squeeze():
    from x_mlps_pytorch.grouped_mlp import GroupedMLP

    mlp = GroupedMLP(256, 128, 64, squeeze_if_one_group = True)

    x = torch.randn(7, 3, 256)

    assert mlp(x).shape == (7, 3, 64)



from x_mlps_pytorch.mlp import (
    MLP,
    create_mlp
)

from x_mlps_pytorch.ff import (
    Feedforwards
)

from x_mlps_pytorch.nff import (
    nFeedforwards
)

from x_mlps_pytorch.grouped_mlp import (
    GroupedMLP
)

from x_mlps_pytorch.grouped_ff import (
    GroupedFeedforwards
)

from x_mlps_pytorch.grouped_nff import (
    nGroupedFeedforwards
)

from x_mlps_pytorch.noisable import (
    Noisable,
    with_seed
)



import pytest
import torch
param = pytest.mark.parametrize

def test_mlp():
    from x_mlps_pytorch.mlp import MLP

    mlp = MLP(256, 128, 64)

    x = torch.randn(7, 3, 256)

    assert mlp(x).shape == (7, 3, 64)

# with depth

def test_create_mlp():
    from x_mlps_pytorch.mlp import create_mlp

    mlp = create_mlp(
        dim = 128,
        dim_in = 256,
        dim_out = 64,
        depth = 4
    )

    # same as MLP(256, 128, 128, 128, 128, 64)

    x = torch.randn(7, 3, 256)

    assert mlp(x).shape == (7, 3, 64)

@param('latent_mlp', (False, True))
@param('condition_hadamard_hiddens', (False, True))
def test_latent_conditioned_mlp(
    latent_mlp,
    condition_hadamard_hiddens
):
    from x_mlps_pytorch.mlp_with_latent import create_mlp

    mlp = create_mlp(256, 4, dim_in = 128, dim_out = 128, dim_latent = 33, latent_mlp = latent_mlp, condition_hiddens = condition_hadamard_hiddens)

    x = torch.randn(7, 3, 128)
    latent = torch.randn(7, 33)

    assert mlp(x, latent = latent).shape == x.shape

@param('rmsnorm', (False, True))
def test_mlp_with_norms(
    rmsnorm
):
    from x_mlps_pytorch.normed_mlp import MLP

    mlp = MLP(256, 128, 64, use_rmsnorm = rmsnorm)

    x = torch.randn(7, 3, 256)

    assert mlp(x).shape == (7, 3, 64)

def test_mlp_list_tensor_input():
    from x_mlps_pytorch.mlp import MLP

    mlp = MLP(256, 128, 64)

    x = [
        torch.randn(7, 3, 128),
        torch.randn(7, 3, 64),
        torch.randn(7, 3, 64),
    ]

    assert mlp(x).shape == (7, 3, 64)



import torch
from torch import nn
import torch.nn.functional as F
from torch.nn import Module, ModuleList

from einops import rearrange, repeat, pack, unpack

# functions

def exists(v):
    return v is not None

def first(arr):
    return arr[0]

def pack_with_inverse(t, pattern):
    packed, shape = pack([t], pattern)

    def inverse(out):
        return first(unpack(out, shape, pattern))

    return packed, inverse

# modules

class GroupedRMSNorm(Module):
    def __init__(
        self,
        dim,
        groups = 1
    ):
        super().__init__()
        self.groups = groups
        self.scale = dim ** 0.5
        self.gamma = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        # grouped l2norm

        x = rearrange(x, '... (g d) n -> ... g d n ', g = self.groups)
        x = F.normalize(x, dim = -2, p = 2)
        x = rearrange(x, '... g d n -> ... (g d) n')

        gamma = rearrange(self.gamma, 'd -> d 1') # channel first

        return gamma * x * self.scale

# main class

class GroupedFeedforwards(Module):

    def __init__(
        self,
        dim,
        depth,
        *,
        dim_in = None,
        dim_out = None,
        activation = nn.GELU(),
        bias = True,
        expansion_factor = 4.,
        final_norm = False,
        groups = 1,
        squeeze_if_one_group = False
    ):
        super().__init__()

        layers = []

        # take care of groups

        self.groups = groups
        self.squeeze_if_one_group = squeeze_if_one_group

        dim = dim * groups
        first_dim = dim * groups

        if exists(dim_in):
            dim_in *= groups
            first_dim = dim_in

        if exists(dim_out):
            dim_out *= groups

        dim_hidden = int(dim * expansion_factor)

        self.first_dim = first_dim

        # layers

        for _ in range(depth):

            layer = nn.Sequential(
                GroupedRMSNorm(dim, groups = groups),
                nn.Conv1d(dim, dim_hidden, 1, bias = bias, groups = groups),
                activation,
                nn.Conv1d(dim_hidden, dim, 1, bias = bias, groups = groups)
            )

            layers.append(layer)

        self.layers = ModuleList(layers)

        # maybe final norm

        self.norm = GroupedRMSNorm(dim, groups = groups) if final_norm else nn.Identity()

        # proj in and out

        self.proj_in = nn.Conv1d(dim_in, dim, 1, groups = groups) if exists(dim_in) else nn.Identity()
        self.proj_out = nn.Conv1d(dim, dim_out, 1, groups = groups) if exists(dim_out) else nn.Identity()

    def forward(
        self,
        x
    ):

        dim = x.shape[-1]

        # channel first

        x = rearrange(x, 'b ... d -> b d ...')

        # repeat for groups if needed

        if dim != self.first_dim:
            x = repeat(x, 'b d ... -> b (g d) ...', g = self.groups)

        # pack

        x, inv_pack = pack_with_inverse(x, 'b d *')

        # project in

        x = self.proj_in(x)

        for layer in self.layers:
            x = layer(x) + x

        x = self.norm(x)

        x = self.proj_out(x)

        # get back the spatial dimensions

        x = inv_pack(x)
        x = rearrange(x, 'b d ... -> b ... d')

        x = rearrange(x, 'b ... (g d) -> b ... g d', g = self.groups)

        if self.squeeze_if_one_group and self.groups == 1:
            x = rearrange(x, 'b ... 1 d -> b ... d')

        return x



import pytest
import torch
param = pytest.mark.parametrize

def test_noisable():
    from torch import nn
    from random import randrange

    from x_mlps_pytorch.noisable import Noisable

    # first define module, wrap with Noisable, make sure forward is the same

    linear = nn.Linear(32, 64)

    noisable_linear = Noisable(linear, noise_scale = 1e-2)

    x = torch.randn(3, 32)

    out = linear(x)
    not_noised_out = noisable_linear(x)

    assert out.shape == (3, 64)
    assert torch.allclose(out, not_noised_out)

    # make sure can noise by passing in {param_name: noise}

    noised_out = noisable_linear(x, noise_for_params = dict(weight = torch.randn(64, 32)))

    assert not torch.allclose(out, noised_out)

    # make sure can noise by passing in {param_name: seed}

    seed = randrange(int(1e7))
    noised_out = noisable_linear(x, noise_for_params = dict(weight = seed))

    assert not torch.allclose(out, noised_out)

    noised_out_again = noisable_linear(x, noise_for_params = dict(weight = seed))

    # able to temporarily noise the weights of a network, for rolling out against environment in memory efficient manner

    with noisable_linear.temp_add_noise_(dict(weight = (seed, 1e-4))):
        assert not torch.allclose(out, noisable_linear(x))

    assert torch.allclose(out, noisable_linear(x))



from __future__ import annotations
from copy import deepcopy
from functools import wraps
from contextlib import contextmanager

import torch
from torch import is_tensor
from torch.nn import Module
from torch.func import functional_call

has_cuda = torch.cuda.is_available()

# helper functions

def exists(v):
    return v is not None

def is_empty(arr):
    return len(arr) == 0

def default(v, d):
    return v if exists(v) else d

# temporary seed

def with_seed(seed):
    def decorator(fn):

        # if no seed, just return the function

        if not exists(seed):
            return fn

        @wraps(fn)
        def inner(*args, **kwargs):
            orig_torch_state = torch.get_rng_state()

            orig_cuda_states = None
            if has_cuda:
                orig_cuda_states = torch.cuda.get_rng_state_all()

            torch.manual_seed(seed)

            if has_cuda:
                torch.cuda.manual_seed_all(seed)

            try:
                out = fn(*args, **kwargs)

            finally:
                torch.set_rng_state(orig_torch_state)

                if has_cuda and orig_cuda_states:
                    torch.cuda.set_rng_state_all(orig_cuda_states)

            return out
        return inner

    return decorator

# wrapper

class Noisable(Module):
    def __init__(
        self,
        model: Module,
        noise_scale = 1.,
        overridable_noise_scale = True
    ):
        super().__init__()
        assert not is_empty(list(model.parameters()))

        self.model = model
        self.noise_scale = noise_scale

        self.overridable_noise_scale = overridable_noise_scale # if set to True, if a noise scale is given alongside a seed in a tuple (int, float), that value replaces the noise scale rather than scales it again

    @property
    def device(self):
        return next(self.model.parameters()).device

    @contextmanager
    def temp_add_noise_(
        self,
        noise_for_params = dict(),
        noise_scale = None,
    ):
        self.get_noised_params(noise_for_params, noise_scale = noise_scale, inplace = True)

        yield

        self.get_noised_params(noise_for_params, noise_scale = noise_scale, inplace = True, negate = True)

    def add_noise_(
        self,
        noise_for_params = dict(),
        noise_scale = None,
        add_to_grad = False,
        negate = False
    ):
        self.get_noised_params(noise_for_params, noise_scale = noise_scale, inplace = True, add_to_grad = add_to_grad, negate = negate)

    def get_noised_params(
        self,
        noise_for_params = dict(),
        inplace = False,
        noise_scale = None,
        negate = False,
        add_to_grad = False
    ):
        # get named params

        named_params = dict(self.model.named_parameters())

        # noise the params

        if not inplace:
            noised_params = deepcopy(named_params)
            return_params = noised_params
        else:
            return_params = named_params

        for name, param in named_params.items():

            param_shape = param.shape

            noise_or_seed = noise_for_params.get(name, None)
            param_noise_scale = default(noise_scale, self.noise_scale)

            if not exists(noise_or_seed):
                continue

            # determine the noise

            if isinstance(noise_or_seed, int):
                noise = with_seed(noise_or_seed)(torch.randn)(param_shape)

            elif isinstance(noise_or_seed, tuple) and len(noise_or_seed) == 2:

                seed, noise_scale_with_seed = noise_or_seed
                noise = with_seed(seed)(torch.randn)(param_shape)

                # maybe overriding noise scale per param

                if self.overridable_noise_scale:
                    param_noise_scale = noise_scale_with_seed
                else:
                    param_noise_scale *= noise_scale_with_seed

            elif is_tensor(noise_or_seed):
                noise = noise_or_seed
            else:
                raise ValueError('invalid type, noise must be float tensor or int')

            noise = noise.to(self.device)

            # scale the noise

            if negate:
                param_noise_scale *= -1

            if param_noise_scale != 1.:
                noise = noise * param_noise_scale

            # if inplace, add directly to param, else set the new dictionary and return that

            if inplace and not add_to_grad:
                # adding noise inplace to params

                param.data.add_(noise)

            elif inplace and add_to_grad:
                # adding noise inplace to grads

                if not exists(param.grad):
                    param.grad = torch.zeros_like(param.data)

                param.grad.add_(noise)

            else:
                # adding to a new dictionary
                noised_params[name] = param + noise

        return return_params

    def forward(
        self,
        *args,
        noise_for_params = dict(),
        noise_scale = None,
        **kwargs
    ):
        if is_empty(noise_for_params):
            return self.model(*args, **kwargs)

        noised_params = self.get_noised_params(noise_for_params, noise_scale = noise_scale)

        # use functional call with noised params

        return functional_call(self.model, noised_params, args, kwargs)
